#!/usr/bin/env python3
######################################################################
# Software License Agreement (BSD License)
#
#  Copyright (c) 2018, Dataspeed Inc.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of Dataspeed Inc. nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.
######################################################################

# Name: gpio_node.py
# Purpose: Provide simple access to analog and digital values and configuration of
#   Numato's usb gpio devices
# TODO: make this documentation follow some sort of python standard
# Reference sites:
# https://numato.com/docs/8-channel-usb-gpio-module-with-analog-inputs/
# https://numato.com/docs/16-channel-usb-gpio-module-with-analog-inputs/
# https://numato.com/docs/32-channel-usb-gpio-module-with-analog-inputs/
# https://numato.com/docs/64-channel-usb-gpio-module-analog-inputs/

import rospy
import serial
import threading

from std_msgs.msg import Float32
from std_msgs.msg import UInt64
from std_msgs.msg import String

lock = threading.Lock()

# wrap serial calls to handle serial exceptions attempting to reopen serial upon failure
# Tested by unplugging and plugging in gpio_usb device while node is running
def ser_exec(f, *args):
   while lock.locked():
       pass
   with lock:
        try:
            if ser_exec.ser == None:
                ser_exec.ser = serial.Serial(node.dev, 19200, timeout=0.01)
                rospy.loginfo("opened gpio serial port")
                node.io_dir_set = False # force I/O direction to be reconfigured, since board was likely a reset
        except serial.SerialException as msg:
            rospy.logerr_throttle(5, "error opening serial port: %s " % (msg))
            return None
        try:
            return f(ser_exec.ser, *args)
        except:
            rospy.logerr_throttle(1, "error using serial port, was the device removed?")
            ser_exec.ser = None # remember that we have to reopen this next time
            return None
ser_exec.ser = None

def _read_analog(ser, ch):
    ser.write(("adc read %d\r" % ch).encode())
    resp = ser.read(25)
    return float(resp[10:-3])/1024.0

def read_analog(ch):
    return ser_exec(_read_analog, ch)

def _read_digital(ser):
    ser.write(("gpio readall\r").encode())
    resp = ser.read(64)
    return int(resp[12:-3], 16)

def read_digital():
    return ser_exec(_read_digital)

def _write_digital(ser, state, pin):
    HIGH = 1
    LOW = 0
    if state == HIGH:
        ser.write(("gpio set %s\r" % str(pin)).encode())
    else:
        ser.write(("gpio clear %s\r" % str(pin)).encode())

def write_digital(state, pin):
    return ser_exec(_write_digital, state, pin)

def _set_io_dir(ser):
    ser.write(("gpio iodir %s\r" % (node.io_dir)).encode())
    ser.read(64)

def set_io_dir():
    ser_exec(_set_io_dir)
    
def output_channels():
    OUT = '0'
    return [ i for (i, dir) in enumerate(node.io_bits) if dir == OUT ]

def input_channels():
    IN = '1'
    return [ i for (i, dir) in enumerate(node.io_bits) if dir == IN ]

#Send "gpio read" command. 
#GPIO number 10 and beyond are referenced in the command by using alphabets starting A. 
#For example GPIO10 will be A, GPIO11 will be B and so on. 
#Please note that this is not intended to be hexadecimal notation 
#so the the alphabets can go beyond F.
#format - pin hi/low - "01", "70", "71" , "R1" , "T1" etc.
def outputCallback(data):
    if len(data.data) == 2:
        gpioIndex = int(data.data[0])
        state = int(data.data[1])
    elif len(data.data) == 3:
        pin1 = int(data.data[0])  # 2
        pin2 = int(data.data[1])  # 9
        state = int(data.data[2])
        gpioNum = str(pin1) + str(pin2) # 29

        if (int(gpioNum) < 10):
            gpioIndex = str(gpioNum)
        else:
            gpioIndex = chr(55 + int(gpioNum))
    write_digital(state, gpioIndex)

# All of the different boards with (8, 16, 32, 64) I/Os have different analog channel to port mappings
# use the adc_map to handle the different configurations for (8, 16, 32, 64) channel boards
adc_map = {}
adc_map[8] = [ 0, 1, 2, 3, None, None, 4, 5]
adc_map[16] = list(range(7)) + 9 * [None]
adc_map[32] = [None] + list(range(1,8)) + 24 * [None]
adc_map[64] = list(range(32)) + 32 * [None]

def analog_channels():
    bd_adc_map = adc_map[node.num_io] 
    return [ bd_adc_map[ch] for ch in input_channels() if bd_adc_map[ch] != None ]

def node():
    rospy.init_node('gpio')

    node.dev = rospy.get_param('~dev', '/dev/ttyACM0')
    node.num_io = int(rospy.get_param('~num_io', '8'))
    if node.num_io == 8:
        node.io_dir = rospy.get_param('~io_dir', 'FF').lower()
    elif node.num_io == 16:
        node.io_dir = rospy.get_param('~io_dir', 'FFFF').lower()
    elif node.num_io == 32:
        node.io_dir = rospy.get_param('~io_dir', 'FFFFFFFF').lower()
    elif node.num_io == 64:
        node.io_dir = rospy.get_param('~io_dir', 'FFFFFFFFFFFFFFFF').lower()
    else:
        rospy.logfatal("%d is not supported" % node.num_io)
        return None

    if node.num_io != len(node.io_dir) * 4:
        rospy.logfatal("length of io_dir, %d bits (specified in hex), does not match num_io, %d" % (len(node.io_dir)*4, node.num_io))
        return None
    #TODO: use serial
    node.serial_num = rospy.get_param('~serial', '')

    node.io_dir_set = False # force I/O dir to be configured

    node.adc_ref = float(rospy.get_param('~adc_ref', '3.3'))

    # From testing, the max measurement rate in Hz can be determined by:
    #   1000/(10.5 + 10.4*NUM_SUBSCRIBED_TOPICS)
    # where NUM_SUBSCRIBED_TOPICS is the total number of input topics that have
    # at least one subscriber. So if all 6 analogs and the digital input are
    # subscribed to, the max rate is about 12 Hz.
    # This does not take into account any output messages.
    rate_hz = float(rospy.get_param('~rate', '10.0'))
    rate = rospy.Rate(rate_hz)

    # make sure num_io is valid
    # TODO: num_io could potentially be auto discovered
    try:
        adc_map[node.num_io]
    except:
        valid_num_ios = [num_io for (num_io, _) in adc_map.items()]
        rospy.logfatal("%d is not a valid number of I/Os, valid number of I/Os are %s" % (node.num_io, valid_num_ios) )
        return None # TODO: better way to shutdown?
    
    # order the io bits in a left to right config for simple [0 to num_ios-1] indexing
    # this allows us to figure out input and ouput configurations ('0' = output, '1' = input)
    node.io_bits = bin(int(node.io_dir, 16))[::-1][0:node.num_io]
    
    rospy.loginfo("inputs: %s" % input_channels())
    rospy.loginfo("adc_inputs: %s" % analog_channels())
    rospy.loginfo("outputs: %s" % output_channels())
    rospy.loginfo("adc_ref: %sV" % node.adc_ref)
    rospy.loginfo("rate: %sHz" % rate_hz)

    rospy.Subscriber ("output_write", String, outputCallback)

    analog_pubs = [ (ch, rospy.Publisher('analog%d' % ch , Float32, queue_size=10)) for ch in analog_channels() ]
    digital_pub = rospy.Publisher('digital', UInt64, queue_size=10)

    while not rospy.is_shutdown():
        if not node.io_dir_set:
            set_io_dir()
        if digital_pub.get_num_connections():
            ddata = read_digital()
            if ddata is not None:
                digital_pub.publish(ddata)
        for (ch, analog_pub) in analog_pubs:
            if analog_pub.get_num_connections():
                adata = read_analog(ch)
                if adata is not None:
                    analog_pub.publish(adata * node.adc_ref)
        rate.sleep()
    if rospy.is_shutdown():
        ser_exec.ser = None
        return None

if __name__ == '__main__':
    try:
        node()
    except rospy.ROSInterruptException:
        pass
